/*----------------------------------------------------------------------------

  SVG Susie Plugin Core Engine
                                                  Copyright(C) fok@Bay-Side
  Main Module

----------------------------------------------------------------------------*/
#include "stdafx.h"

/*----------------------------------------------------------------------------
	RenderSVG̃tO
----------------------------------------------------------------------------*/
#define CDFSVG_GETIMAGESIZEONLY		0x00000004		// _OABITMAPINFOHEADERo͂ďI
#define CDFSVG_WITHFILEHEADER		0x00000008		// BITMAPFILEHEADERo͂

/*----------------------------------------------------------------------------
	fobOp
----------------------------------------------------------------------------*/
#ifdef _DEBUG
	#define d_printf _cprintf
#else
	#define d_printf __noop
#endif

/*----------------------------------------------------------------------------
	DIBZNV쐬
----------------------------------------------------------------------------*/
static inline HBITMAP MyCreateDIBSection(INT nWidth, INT nHeight, INT nBits, LPVOID *ppvBits)
{
	BITMAPINFOHEADER bmiHeader;
	LPVOID pvBits;
	HBITMAP hbm;
	HDC hdc;

	hdc = CreateCompatibleDC(NULL);

	ZeroMemory(&bmiHeader, sizeof(BITMAPINFOHEADER));

	bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmiHeader.biWidth = nWidth;
	bmiHeader.biHeight = nHeight;
	bmiHeader.biPlanes = 1;
	bmiHeader.biBitCount = nBits;

	hbm = CreateDIBSection(hdc, (LPBITMAPINFO)&bmiHeader, DIB_RGB_COLORS, &pvBits, NULL, 0);

	if(hbm && ppvBits)
	{
		*ppvBits = pvBits;
	}

	DeleteDC(hdc);

	return hbm;
}

/*----------------------------------------------------------------------------
	w`ōő\łTCYvZ
----------------------------------------------------------------------------*/
static inline bool RecalcSizeRect(LPCRECT prcSrcRect, INT nWidth, INT nHeight, RECT* prcResult)
{
	if(!prcSrcRect || !prcResult || !nWidth || !nHeight || !prcResult)
	{
		return false;
	}

	const INT scr_w = prcSrcRect->right - prcSrcRect->left;
	const INT scr_h = prcSrcRect->bottom - prcSrcRect->top;
	INT xsize = nWidth;
	INT ysize = nHeight;

	if((scr_w * ysize) >= (xsize * scr_h))
	{
		// E󂭂
		xsize = xsize * scr_h / ysize;
		ysize = scr_h;

		prcResult->left = prcSrcRect->left + (scr_w - xsize) / 2;
		prcResult->right = prcResult->left + xsize;
		prcResult->top = prcSrcRect->top;
		prcResult->bottom = prcResult->top + ysize;
	}
	else
	{
		// ㉺
		ysize = ysize * scr_w / xsize;
		xsize = scr_w;

		prcResult->left = prcSrcRect->left;
		prcResult->right = prcResult->left + xsize;
		prcResult->top = prcSrcRect->top + (scr_h - ysize) / 2;
		prcResult->bottom = prcResult->top + ysize;
	}

	return true;
}

/*----------------------------------------------------------------------------
	SVGDIB쐬ďo
----------------------------------------------------------------------------*/
static bool RenderSVG(HANDLE hOutput, LPCSTR lpszSrcFileName, DWORD dwFlags, INT nBitCount)
{
	RsvgHandle *r_svg = NULL;
	RsvgDimensionData svgdim;
	cairo_surface_t* surface = NULL;
	cairo_t* cr = NULL;
	HBITMAP hbmDIB = NULL;
	HBITMAP hbmPrev = NULL;
	HDC hdc = NULL;

	// \[Xމ
	BOOST_SCOPE_EXIT((&r_svg)(&surface)(&cr)(&hbmDIB)(&hbmPrev)(&hdc))
	{
		if(cr)
		{
			cairo_destroy(cr);
		}
		if(surface)
		{
			cairo_surface_destroy(surface);
		}
		if(r_svg)
		{
			rsvg_handle_free(r_svg);
		}
		if(hbmPrev)
		{
			SelectObject(hdc, hbmPrev);
		}
		if(hbmDIB)
		{
			DeleteObject(hbmDIB);
		}
		if(hdc)
		{
			DeleteDC(hdc);
		}
	}
	BOOST_SCOPE_EXIT_END;

	// SVGǂݍ
	d_printf("rsvg_handle_new_from_file begin, %s\n", lpszSrcFileName);
	r_svg = rsvg_handle_new_from_file(lpszSrcFileName, NULL);
	d_printf("rsvg_handle_new_from_file end\n");
	if(!r_svg)
	{
		return false;
	}

	// SVGf[^̏擾
	d_printf("rsvg_handle_get_dimensions begin\n");
	rsvg_handle_get_dimensions(r_svg, &svgdim);
	d_printf("rsvg_handle_get_dimensions end\n");
	DWORD dwSrcWidth = abs(svgdim.width);
	DWORD dwSrcHeight = abs(svgdim.height);
	DWORD dwRenderWidth = dwSrcWidth;
	DWORD dwRenderHeight = dwSrcHeight;

	// DIBwb_
	DWORD dwWrite;
	BITMAPINFOHEADER bmi;
	ZeroMemory(&bmi, sizeof(bmi));
	bmi.biSize = sizeof(bmi);
	bmi.biWidth = dwSrcWidth;
	bmi.biHeight = dwSrcHeight;
	bmi.biPlanes = 1;
	bmi.biBitCount = nBitCount;
	bmi.biSizeImage = (((bmi.biWidth * bmi.biBitCount) + 31) / 32 * 4) * bmi.biHeight;

	// BITMAPFILEHEADEȐo͂邩H
	if(dwFlags & CDFSVG_WITHFILEHEADER)
	{
		BITMAPFILEHEADER bmf;
		ZeroMemory(&bmf, sizeof(bmf));
		bmf.bfType = MAKEWORD('B', 'M');
		bmf.bfSize = sizeof(bmf) + sizeof(bmi) + bmi.biSizeImage;
		bmf.bfOffBits = sizeof(bmf) + sizeof(bmi);

		d_printf("output BITMAPFILEHEADER begin\n");
		dwWrite = 0;
		if(!WriteFile(hOutput, &bmf, sizeof(bmf), &dwWrite, NULL))
		{
			d_printf("output BITMAPFILEHEADER NG 1\n");
			return false;
		}
		if(dwWrite != sizeof(bmf))
		{
			d_printf("output BITMAPFILEHEADER NG 2\n");
			return false;
		}
		d_printf("output BITMAPFILEHEADER OK, written = %d\n", dwWrite);
	}

	// DIBwb_̏o
	d_printf("output BITMAPINFOHEADER begin\n");
	dwWrite = 0;
	if(!WriteFile(hOutput, &bmi, sizeof(bmi), &dwWrite, NULL))
	{
		d_printf("output BITMAPINFOHEADER NG 1\n");
		return false;
	}
	if(dwWrite != sizeof(bmi))
	{
		d_printf("output BITMAPINFOHEADER NG 2\n");
		return false;
	}
	d_printf("output BITMAPINFOHEADER OK, written = %d\n", dwWrite);

	// 擾݂̂ł΂ŏI
	if(dwFlags & CDFSVG_GETIMAGESIZEONLY)
	{
		d_printf("CDFSVG_GETIMAGESIZEONLY return\n");
		// ŏI
		return true;
	}

	// o͐DIBZNVƃfoCXReLXg쐬
	d_printf("CreateCompatibleDC()\n");
	hdc = CreateCompatibleDC(NULL);
	if(!hdc)
	{
		return false;
	}
	LPVOID lpBuffer = NULL;
	hbmDIB = MyCreateDIBSection(dwRenderWidth, dwRenderHeight, bmi.biBitCount, &lpBuffer);
	if(!hbmDIB)
	{
		return false;
	}
	d_printf("SelectObject()\n");
	hbmPrev = (HBITMAP)SelectObject(hdc, hbmDIB);
	if(bmi.biBitCount != 32)
	{
		// At@`l̏ꍇ͔œhԂĂ
		d_printf("BitBlt()\n");
		BitBlt(hdc, 0, 0, dwRenderWidth, dwRenderHeight, hdc, 0, 0, WHITENESS);
	}

	// CairoT[tFX쐬
	d_printf("cairo_win32_surface_create begin\n");
	surface = cairo_win32_surface_create(hdc);
	d_printf("cairo_win32_surface_create end\n");
	if(!surface)
	{
		return false;
	}
	d_printf("cairo_create begin\n");
	cr = cairo_create(surface);
	d_printf("cairo_create end\n");
	if(!cr)
	{
		return false;
	}

	// XP[Ow͕KvH
	double dbScaleW = 1.0, dbScaleH = 1.0;
	if(dwRenderWidth != dwSrcWidth)
	{
		dbScaleW = double(dwRenderWidth) / double(dwSrcWidth);
	}
	if(dwRenderHeight != dwSrcHeight)
	{
		dbScaleH = double(dwRenderHeight) / double(dwSrcHeight);
	}
	if(dbScaleW != 1.0 || dbScaleH != 1.0)
	{
		// XP[Ow
		cairo_matrix_t matrix;
		cairo_matrix_init_identity(&matrix);
		cairo_matrix_scale(&matrix, dbScaleW, dbScaleH);
		cairo_transform(cr, &matrix);
	}

	// _Os
	d_printf("rsvg_handle_render_cairo begin\n");
	if(rsvg_handle_render_cairo(r_svg, cr) == 0)
	{
		d_printf("rsvg_handle_render_cairo ERROR\n");
		return false;
	}
	d_printf("rsvg_handle_render_cairo OK\n");

	// rbgf[^o
	GdiFlush();
	d_printf("output bits begin\n");
	dwWrite = 0;
	if(!WriteFile(hOutput, lpBuffer, bmi.biSizeImage, &dwWrite, NULL))
	{
		d_printf("output bits NG 1\n");
		return false;
	}
	if(dwWrite != bmi.biSizeImage)
	{
		d_printf("output bits NG 2\n");
		return false;
	}
	d_printf("output bits OK, written = %d\n", dwWrite);

	return true;
}

/*----------------------------------------------------------------------------
	CGg|Cg
----------------------------------------------------------------------------*/
INT _tmain(INT argc, TCHAR* argv[])
{
	if(argc < 2)
	{
		return -1;
	}

	TCHAR szFileName[MAX_PATH];
	DWORD dwFlags = 0;
	ZeroMemory(szFileName, sizeof(szFileName));
	INT i = 1;
	INT nBitCount = 24;

	// XCb``FbN
	if(argc > 2) // t@CK{Ȃ̂argc3ȏłȂƃXCb`ӖȂ
	{
		if(0 == lstrcmpi(argv[i], _T("/getinfo")))
		{
			// wb_̂
			dwFlags |= CDFSVG_GETIMAGESIZEONLY;
			i++;
		}
		else if(0 == lstrcmpi(argv[i], _T("/bmpheader")))
		{
			// BITMAPFILEHEADERo
			dwFlags |= CDFSVG_WITHFILEHEADER;
			i++;
		}

		if(argc > (i + 1))
		{
			if(0 == lstrcmpi(argv[i], _T("/32bit")))
			{
				// 32bitDIBŏo
				nBitCount = 32;
				i++;
			}
		}
	}

	if(argc > i)
	{
		// t@C݂Ȃ΃G[
		_tcscpy_s(szFileName, argv[i]);
		if(!PathFileExists(szFileName))
		{
			return -2;
		}
	}
	else
	{
		return -3;
	}

	// _Os
	rsvg_init();
	bool bRet = RenderSVG(GetStdHandle(STD_OUTPUT_HANDLE), CT2A(szFileName), dwFlags, nBitCount);
	rsvg_term();
	return bRet ? 0 : 1;
}
